// Complex.h
//
// ICS 65 Fall 2012
// Code Example
//
// This header file declares a Complex class, a simple (and intentionally
// incomplete) implementation of a complex number type.  It overloads
// at least some of the operators that you might think of as sensible,
// while also providing other features that make sense, such as implicit
// type conversions to and from the built-in numeric types.

#ifndef COMPLEX_H
#define COMPLEX_H

#include <iostream>


// A Complex object represents a complex number (a + bi) from mathematics,
// where a is the "real part" and bi is the "imaginary part".

class Complex
{
public:
	// I've opted not to make this constructor explicit, even though it can
	// be called with one argument.  The reason is because I think this
	// line of code (and many others like it) should be legal, since real
	// numbers are also complex numbers (with zero as the imaginary part).
	//
	//     Complex c = 3.5;
    //
	// I figure that should set c to have a real part of 3.5 and an
    // imaginary part of 0.0, which is exactly what this constructor
    // will do in the situation above.
	Complex(double initialRealPart, double initialImaginaryPart = 0.0)
		: realPart(initialRealPart), imaginaryPart(initialImaginaryPart)
	{
	}
	
	// To give Complex numbers the same performance characteristics as
	// built-in numeric types (i.e., they don't get initialized automatically
	// if not explicitly assigned an initial value), I'll include a default
	// constructor that does no initialization.
	Complex() { }


	// I'm not going to bother defining an assignment operator or a copy
	// constructor, because the default versions of each will do exactly
	// the right thing.  Remember that the default copy constructor makes
	// a copy of each member into the new object; since the members are
	// doubles (and not pointers to dynamically-allocated objects), this
	// behavior is fine.  Similarly, the default assignment operator
	// does a memberwise assignment of one object into another, effectively
	// copying the two doubles into the target object.  That's fine, too.


	// Rather than defining separate accessors and mutators, like getReal()
	// and setReal(), I'm going to opt for a simpler approach that will
	// yield faster code: implement member functions that return references
	// to the realPart and imaginaryPart member variables so that they can
	// be used either on the left or right side of an assignment statement.
	// For example:
	//
	//     Complex c;
	//     cout << c.real() << endl;
	//     c.real() = 5.5;
	//
	// Note that there are two separate versions of each function, one that's
	// const and one that isn't.  This is because we'd like to restrict
	// the use of the function as a mutator if the Complex it's called on is
	// const.  So we overload the function on the basis of its constness.
	// The non-const one will be called on a non-const Complex; the const one
	// will be called on a const Complex.  Note that the const version returns
	// a const reference to the realPart, meaning that it can be accessed but
	// not changed.
	
	double& real() { return realPart; }
	const double& real() const { return realPart; }
	
	double& imaginary() { return imaginaryPart; }
	const double& imaginary() const { return imaginaryPart; }


	// These are the declarations of the += and -= operators.  It may seem
	// odd that these functions take only one parameter, but remember that
	// member functions have an additional parameter -- "this" -- that is
	// not listed.  In the case of these operators, "this" refers to the
	// object on the left-hand side of the += operation, while "d" (the
	// explicitly listed parameter) refers to the object on the right-hand
	// side.
	Complex& operator+=(const Complex& d);
	Complex& operator-=(const Complex& d);

private:
	double realPart;
	double imaginaryPart;
};


// Because accessor/mutator functions are provided for the real and
// imaginary parts, it turns out that none of the operators need to be
// member functions (or even friends).
//
// Remember the benefit... By making your operators *not* be members, you
// allow them to be used the most flexibly.  For example, since the +
// operator is not a member, any of these expressions will be legal, given
// two Complex variables c and d:
//
//     c + d
//     c + 3.0
//     3.0 + d
//
// If the + operator is a member function within Complex, the third of
// those expressions (3.0 + d) will no longer be legal.  The reason is
// that the first (implicit) argument to a Complex member function (operator
// or otherwise) must be a Complex object, and not something that's implicitly
// converted into a Complex before the function call is made.  This is the
// same reason why this code wouldn't be legal either:
//
//     double d = 3.5;
//     d.imaginary() = 5.0;
//
// Even though doubles are implicitly convertible to Complex objects, C++
// doesn't convert an object implicitly before calling a member function
// on it, since it would lead to legality for some pretty bizarre code, like
// the code above.
//
// While we should favor having our operators not be member functions in
// cases when implicit conversions are legal, the += and -= operators are
// listed as member functions.  These operators will need to modify the
// Complex on the left-hand side, e.g., in the expression c += 3.0 or c += d.
// Since the left-hand side must always be a Complex, it's sensible to make
// these operators be member functions and implement them using direct access
// to fields, since they'll likely inline better.
//
// Now, another solution to this would be to make all of these operator
// functions be friends of the Complex class, but I've opted against that
// because none of them need direct access to Complex's members.  Limiting
// the number of functions that are members or friends is generally a good
// thing; friendship used conservatively benefits a design, but using it
// too much causes the design to be much more difficult to understand.
//
// Since I want code in other source files to use these operators, I'm
// declaring the functions in the header file.  I'll define the functions
// in Complex.cpp.

Complex operator+(const Complex& c, const Complex& d);
Complex operator-(const Complex& c, const Complex& d);
bool operator==(const Complex& c, const Complex& d);
bool operator!=(const Complex& c, const Complex& d);
std::ostream& operator<<(std::ostream& out, const Complex& c);
std::istream& operator>>(std::istream& in, Complex& c);


#endif // COMPLEX_H
